@@ -23,7 +23,6 @@ import sys
 
 from . import keywords
 from . import tokenize
-from . import utils
 
 
 try:
@@ -44,7 +43,7 @@ __author__ = 'nnorwitz@google.com (Neal Norwitz)'
 #
 # TODO much, much later:
 # * Handle #define
-#  * exceptions
+# * exceptions
 
 
 VISIBILITY_PUBLIC, VISIBILITY_PROTECTED, VISIBILITY_PRIVATE = list(range(3))
@@ -89,13 +88,7 @@ class Node(object):
         """Returns bool if this node exportable from a header file."""
         return False
 
-    def requires(self, node):
-        """Does this AST node require the definition of the node passed in?"""
-        return False
-
     def _string_helper(self, name, suffix):
-        if not utils.DEBUG:
-            return '%s(%s)' % (name, suffix)
         return '%s(%d, %d, %s)' % (name, self.start, self.end, suffix)
 
     def __repr__(self):
@@ -146,10 +139,6 @@ class Expr(Node):
         Node.__init__(self, start, end)
         self.expr = expr
 
-    def requires(self, node):
-        # TODO(nnorwitz): impl.
-        return False
-
     def __str__(self):
         return self._string_helper(self.__class__.__name__, unicode(self.expr))
 
@@ -188,10 +177,6 @@ class Parameter(Node):
         self.type = parameter_type
         self.default = default
 
-    def requires(self, node):
-        # TODO(nnorwitz): handle namespaces, etc.
-        return self.type.name == node.name
-
     def __str__(self):
         name = unicode(self.type)
         suffix = '%s %s' % (name, self.name)
@@ -228,10 +213,6 @@ class VariableDeclaration(_GenericDeclaration):
         self.type = var_type
         self.initial_value = initial_value
 
-    def requires(self, node):
-        # TODO(nnorwitz): handle namespaces, etc.
-        return self.type.name == node.name
-
     def to_string(self):
         """Return a string that tries to reconstitute the variable decl."""
         suffix = '%s %s' % (self.type, self.name)
@@ -255,14 +236,6 @@ class Typedef(_GenericDeclaration):
     def is_exportable(self):
         return True
 
-    def requires(self, node):
-        # TODO(nnorwitz): handle namespaces, etc.
-        name = node.name
-        for token in self.alias:
-            if token is not None and name == token.name:
-                return True
-        return False
-
     def __str__(self):
         suffix = '%s, %s' % (self.name, self.alias)
         return self._type_string_helper(suffix)
@@ -311,17 +284,6 @@ class Class(_GenericDeclaration):
     def is_exportable(self):
         return not self.is_declaration()
 
-    def requires(self, node):
-        # TODO(nnorwitz): handle namespaces, etc.
-        if self.bases:
-            for token_list in self.bases:
-                # TODO(nnorwitz): bases are tokens, do name comparison.
-                for token in token_list:
-                    if token.name == node.name:
-                        return True
-        # TODO(nnorwitz): search in body too.
-        return False
-
     def __str__(self):
         name = self.name
         if self.templated_types:
@@ -357,15 +319,6 @@ class Function(_GenericDeclaration):
             return False
         return None not in self.namespace
 
-    def requires(self, node):
-        if self.parameters:
-            # TODO(nnorwitz): parameters are tokens, do name comparison.
-            for p in self.parameters:
-                if p.name == node.name:
-                    return True
-        # TODO(nnorwitz): search in body too.
-        return False
-
     def __str__(self):
         # TODO(nnorwitz): add templated_types.
         suffix = ('%s %s(%s), 0x%02x, %s' %
@@ -482,6 +435,9 @@ class TypeConverter(object):
                 else:
                     names.append(t.name)
             name = ''.join(names)
+
+            assert_parse(name_tokens, 'not enough tokens')
+
             result.append(Type(name_tokens[0].start, name_tokens[-1].end,
                                name, templated_types, modifiers,
                                reference, pointer, array))
@@ -493,6 +449,14 @@ class TypeConverter(object):
             token = tokens[i]
             if token.name == '<':
                 new_tokens, new_end = self._get_template_end(tokens, i + 1)
+                if new_end < end:
+                    if tokens[new_end].name == '::':
+                        name_tokens.append(tokens[new_end])
+                        name_tokens.append(tokens[new_end + 1])
+                        new_end += 2
+                    elif tokens[new_end].name == '*':
+                        pointer = True
+                        new_end += 1
                 add_type(self.to_type(new_tokens))
                 # If there is a comma after the template, we need to consume
                 # that here otherwise it becomes part of the name.
@@ -601,10 +565,19 @@ class TypeConverter(object):
                 del default[0]  # Remove flag.
             end = type_modifiers[-1].end
 
-            needs_name_removed = not (
-                len(type_modifiers) == 1 or
-                (len(type_modifiers) == 2 and
-                 type_modifiers[0].name == 'const'))
+            needs_name_removed = True
+            if len(type_modifiers) == 1:
+                needs_name_removed = False
+            else:
+                last = type_modifiers[-1].name
+                second_to_last = type_modifiers[-2].name
+                if (
+                    last == '>' or
+                    keywords.is_builtin_type(last) or
+                    second_to_last == '::' or
+                    keywords.is_builtin_type_modifiers(second_to_last)
+                ):
+                    needs_name_removed = False
 
             (name, type_name, templated_types, modifiers,
              _, __) = self.declaration_to_parts(type_modifiers,
@@ -626,7 +599,10 @@ class TypeConverter(object):
             elif s.name == '>':
                 template_count -= 1
             if template_count > 0:
-                type_modifiers.append(s)
+                if default:
+                    default.append(s)
+                else:
+                    type_modifiers.append(s)
                 continue
 
             if s.name == ',':
@@ -721,9 +697,11 @@ class AstBuilder(object):
             self.current_token = token
 
             # Dispatch on the next token type.
-            if token.token_type == _INTERNAL_TOKEN:
-                if token.name == _NAMESPACE_POP:
-                    self.namespace_stack.pop()
+            if (
+                token.token_type == _INTERNAL_TOKEN and
+                token.name == _NAMESPACE_POP
+            ):
+                self.namespace_stack.pop()
                 continue
 
             try:
@@ -770,7 +748,10 @@ class AstBuilder(object):
             # Handle data or function declaration/definition.
             syntax = tokenize.SYNTAX
             temp_tokens, last_token = \
-                self._get_var_tokens_up_to(syntax, '(', ';', '{', '[')
+                self._get_var_tokens_up_to(syntax, '(', ';', '{', '[', '}')
+            if last_token.name == '}':
+                return None
+
             temp_tokens.insert(0, token)
             if last_token.name == '(':
                 # If there is an assignment before the paren,
@@ -780,10 +761,10 @@ class AstBuilder(object):
                         elt.name == '=' and
                         temp_tokens[i - 1].name != 'operator'
                     ):
-                        new_temp = self._get_tokens_up_to(tokenize.SYNTAX, ';')
                         temp_tokens.append(last_token)
+                        new_temp, last_token = \
+                            self._get_var_tokens_up_to(tokenize.SYNTAX, ';')
                         temp_tokens.extend(new_temp)
-                        last_token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
                         break
 
             if last_token.name == '[':
@@ -802,8 +783,7 @@ class AstBuilder(object):
                 name, type_name, templated_types, modifiers, default, _ = \
                     self.converter.declaration_to_parts(temp_tokens, True)
 
-                if not temp_tokens:
-                    raise ParseError('not enough tokens')
+                assert_parse(temp_tokens, 'not enough tokens')
 
                 t0 = temp_tokens[0]
                 names = [t.name for t in temp_tokens]
@@ -814,6 +794,8 @@ class AstBuilder(object):
                 return self._create_variable(t0, name, type_name, modifiers,
                                              names, templated_types, default)
             if last_token.name == '{':
+                assert_parse(temp_tokens, 'not enough tokens')
+
                 self._add_back_tokens(temp_tokens[1:])
                 self._add_back_token(last_token)
                 method_name = temp_tokens[0].name
@@ -845,8 +827,10 @@ class AstBuilder(object):
                 # Handle #include \<newline> "header-on-second-line.h".
                 if name.startswith('\\'):
                     name = name[1:].strip()
-                if name[0] not in '<"' or name[-1] not in '>"':
-                    raise ParseError(token)
+
+                assert_parse(name[0] in '<"', token)
+                assert_parse(name[-1] in '>"', token)
+
                 system = name[0] == '<'
                 filename = name[1:-1]
                 return Include(token.start, token.end, filename, system)
@@ -880,7 +864,6 @@ class AstBuilder(object):
             last_token = self._get_next_token()
         return tokens, last_token
 
-    # TODO(nnorwitz): remove _ignore_up_to() it shouldn't be necessary.
     def _ignore_up_to(self, token_type, token):
         self._get_tokens_up_to(token_type, token)
 
@@ -930,22 +913,11 @@ class AstBuilder(object):
         return next(self.tokens)
 
     def _add_back_token(self, token):
-        if token.whence == tokenize.WHENCE_STREAM:
-            token.whence = tokenize.WHENCE_QUEUE
-            self.token_queue.insert(0, token)
-        else:
-            assert token.whence == tokenize.WHENCE_QUEUE, token
-            self.token_queue.append(token)
+        self.token_queue.append(token)
 
     def _add_back_tokens(self, tokens):
         if tokens:
-            if tokens[-1].whence == tokenize.WHENCE_STREAM:
-                for token in tokens:
-                    token.whence = tokenize.WHENCE_QUEUE
-                self.token_queue[:0] = reversed(tokens)
-            else:
-                assert tokens[-1].whence == tokenize.WHENCE_QUEUE, tokens
-                self.token_queue.extend(reversed(tokens))
+            self.token_queue.extend(reversed(tokens))
 
     def get_name(self, seq=None):
         """Returns ([tokens], next_token_info)."""
@@ -985,14 +957,14 @@ class AstBuilder(object):
         template_portion = None
         if get_paren:
             token = self._get_next_token()
-            assert token.token_type == tokenize.SYNTAX, token
+            assert_parse(token.token_type == tokenize.SYNTAX, token)
             if token.name == '<':
                 # Handle templatized dtors.
                 template_portion = [token]
                 template_portion.extend(self._get_matching_char('<', '>'))
                 token = self._get_next_token()
-            assert token.token_type == tokenize.SYNTAX, token
-            assert token.name == '(', token
+            assert_parse(token.token_type == tokenize.SYNTAX, token)
+            assert_parse(token.name == '(', token)
 
         name = return_type_and_name.pop()
         if (len(return_type_and_name) > 1 and
@@ -1011,8 +983,8 @@ class AstBuilder(object):
             name = return_type_and_name.pop()
         elif name.name == ']':
             rt = return_type_and_name
-            assert rt[-1].name == '[', return_type_and_name
-            assert rt[-2].name == 'operator', return_type_and_name
+            assert_parse(rt[-1].name == '[', return_type_and_name)
+            assert_parse(rt[-2].name == 'operator', return_type_and_name)
             name_seq = return_type_and_name[-2:]
             del return_type_and_name[-2:]
             name = tokenize.Token(tokenize.NAME, 'operator[]',
@@ -1035,7 +1007,7 @@ class AstBuilder(object):
         # Handling operator() is especially weird.
         if name.name == 'operator' and not parameters:
             token = self._get_next_token()
-            assert token.name == '(', token
+            assert_parse(token.name == '(', token)
             name = tokenize.Token(tokenize.NAME, 'operator()',
                                   name.start, last_token.end)
             parameters = list(self._get_parameters())
@@ -1054,14 +1026,14 @@ class AstBuilder(object):
                 # TODO(nnorwitz): handle more __attribute__ details.
                 modifiers |= FUNCTION_ATTRIBUTE
                 token = self._get_next_token()
-                assert token.name == '(', token
+                assert_parse(token.name == '(', token)
                 # Consume everything between the (parens).
                 list(self._get_matching_char('(', ')'))
                 token = self._get_next_token()
             elif token.name == 'throw':
                 modifiers |= FUNCTION_THROW
                 token = self._get_next_token()
-                assert token.name == '(', token
+                assert_parse(token.name == '(', token)
                 # Consume everything between the (parens).
                 list(self._get_matching_char('(', ')'))
                 token = self._get_next_token()
@@ -1074,8 +1046,16 @@ class AstBuilder(object):
                 self._add_back_token(token)
                 token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
 
-        if token.token_type != tokenize.SYNTAX:
-            raise ParseError(token)
+        # Handle ref-qualifiers.
+        if token.name == '&' or token.name == '&&':
+            token = self._get_next_token()
+
+        if token.name == '}' or token.name == '#endif':
+            self._add_back_token(token)
+            token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
+
+        assert_parse(token.token_type == tokenize.SYNTAX, token)
+
         # Handle ctor initializers.
         if token.name == ':':
             # TODO(nnorwitz): anything else to handle for initializer list?
@@ -1095,8 +1075,10 @@ class AstBuilder(object):
                 del function_parameters[-1]  # Remove trailing ')'.
                 # TODO(nnorwitz): store the function_parameters.
                 token = self._get_next_token()
-                if token.token_type != tokenize.SYNTAX or token.name != ';':
-                    raise ParseError(token)
+
+                assert_parse(token.token_type == tokenize.SYNTAX, token)
+                assert_parse(token.name == ';', token)
+
                 return self._create_variable(indices, name.name, indices.name,
                                              modifiers, '')
             # At this point, we got something like:
@@ -1120,8 +1102,8 @@ class AstBuilder(object):
             body = None
             if token.name == '=':
                 token = self._get_next_token()
-                assert token.token_type == tokenize.CONSTANT, token
-                assert token.name == '0', token
+                assert_parse(token.token_type == tokenize.CONSTANT, token)
+                assert_parse(token.name == '0', token)
                 modifiers |= FUNCTION_PURE_VIRTUAL
                 token = self._get_next_token()
 
@@ -1131,8 +1113,8 @@ class AstBuilder(object):
                 list(self._get_matching_char('[', ']'))
                 token = self._get_next_token()
 
-            if token.name != ';':
-                raise ParseError((token, return_type_and_name, parameters))
+            assert_parse(token.name == ';',
+                         (token, return_type_and_name, parameters))
 
         # Looks like we got a method, not a function.
         if len(return_type) > 1 and return_type[-1].name == '::':
@@ -1172,7 +1154,6 @@ class AstBuilder(object):
         while i < end:
             # Iterate through the sequence parsing out each name.
             new_name, next_item = self.get_name(seq_copy[i:])
-            assert new_name, 'Got empty new_name, next=%s' % next_item
             # We got a pointer or ref. Add it to the name.
             if next_item and next_item.token_type == tokenize.SYNTAX:
                 new_name.append(next_item)
@@ -1223,25 +1204,45 @@ class AstBuilder(object):
         pass
 
     def _get_nested_type(self, ctor):
+        # Handle strongly typed enumerations.
+        token = self._get_next_token()
+        if token.name != 'class':
+            self._add_back_token(token)
+
         name = None
         name_tokens, token = self.get_name()
         if name_tokens:
             name = ''.join([t.name for t in name_tokens])
 
+        if token.token_type == tokenize.NAME:
+            if self._handling_typedef:
+                self._add_back_token(token)
+                return ctor(token.start, token.end, name, None,
+                            self.namespace_stack)
+
+            next_token = self._get_next_token()
+            if next_token.name != '(':
+                self._add_back_token(next_token)
+            else:
+                token = next_token
+
+        if token.token_type == tokenize.SYNTAX and token.name == '(':
+            self._ignore_up_to(tokenize.SYNTAX, ')')
+            token = self._get_next_token()
+
+        # Handle underlying type.
+        if token.token_type == tokenize.SYNTAX and token.name == ':':
+            _, token = self._get_var_tokens_up_to(tokenize.SYNTAX, '{', ';')
+
         # Handle forward declarations.
         if token.token_type == tokenize.SYNTAX and token.name == ';':
             return ctor(token.start, token.end, name, None,
                         self.namespace_stack)
 
-        if token.token_type == tokenize.NAME and self._handling_typedef:
-            self._add_back_token(token)
-            return ctor(token.start, token.end, name, None,
-                        self.namespace_stack)
-
         # Must be the type declaration.
-        fields = list(self._get_matching_char('{', '}'))
-        del fields[-1]                  # Remove trailing '}'.
         if token.token_type == tokenize.SYNTAX and token.name == '{':
+            fields = list(self._get_matching_char('{', '}'))
+            del fields[-1]                  # Remove trailing '}'.
             next_item = self._get_next_token()
             new_type = ctor(token.start, token.end, name, fields,
                             self.namespace_stack)
@@ -1253,7 +1254,7 @@ class AstBuilder(object):
             token = next_item
 
         # Must be variable declaration using the type prefixed with keyword.
-        assert token.token_type == tokenize.NAME, token
+        assert_parse(token.token_type == tokenize.NAME, token)
         return self._create_variable(token, token.name, name, [], '')
 
     def _handle_class_and_struct(self, class_type, class_str, visibility):
@@ -1273,8 +1274,10 @@ class AstBuilder(object):
                            next_token.name == ';')
             variable = var_token
             if is_syntax and not is_variable:
-                variable = next_token
-                temp = self._get_next_token()
+                temp = next_token
+                while variable.token_type != tokenize.NAME:
+                    variable = temp
+                    temp = self._get_next_token()
                 if temp.token_type == tokenize.SYNTAX and temp.name == '(':
                     # Handle methods declared to return a class/struct.
                     t0 = name_tokens[0]
@@ -1284,8 +1287,9 @@ class AstBuilder(object):
                     type_and_name.extend(name_tokens)
                     type_and_name.extend((var_token, next_token))
                     return self._get_method(type_and_name, 0, None, False)
-                if temp.name != ';':
-                    raise ParseError((temp, name_tokens, var_token))
+
+                assert_parse(temp.name == ';', (temp, name_tokens, var_token))
+
             if is_syntax or (is_variable and not self._handling_typedef):
                 modifiers = [class_str]
                 type_name = ''.join([t.name for t in name_tokens])
@@ -1339,7 +1343,8 @@ class AstBuilder(object):
             token2 = self._get_next_token()
         if token2.token_type == tokenize.SYNTAX and token2.name == '~':
             return self.get_method(FUNCTION_VIRTUAL + FUNCTION_DTOR, None)
-        assert token.token_type == tokenize.NAME or token.name == '::', token
+        assert_parse(token.token_type == tokenize.NAME or token.name == '::',
+                     token)
         return_type_and_name = self._get_tokens_up_to(tokenize.SYNTAX, '(')
         return_type_and_name.insert(0, token)
         if token2 is not token:
@@ -1354,8 +1359,7 @@ class AstBuilder(object):
         pass
 
     def handle_public(self):
-        if not self.in_class:
-            raise ParseError('expected to be in a class')
+        assert_parse(self.in_class, 'expected to be in a class')
         self.visibility = VISIBILITY_PUBLIC
 
     def handle_protected(self):
@@ -1452,7 +1456,7 @@ class AstBuilder(object):
             if i < len_tokens:
                 i += 1
                 if tokens[i - 1].name == '=':
-                    assert i < len_tokens, '%s %s' % (i, tokens)
+                    assert_parse(i < len_tokens, '%s %s' % (i, tokens))
                     default, _ = self.get_name(tokens[i:])
                     i += len(default)
                 else:
@@ -1467,8 +1471,10 @@ class AstBuilder(object):
 
     def handle_template(self):
         token = self._get_next_token()
-        assert token.token_type == tokenize.SYNTAX, token
-        assert token.name == '<', token
+
+        assert_parse(token.token_type == tokenize.SYNTAX, token)
+        assert_parse(token.name == '<', token)
+
         templated_types = self._get_templated_types()
         # TODO(nnorwitz): for now, just ignore the template params.
         token = self._get_next_token()
@@ -1506,7 +1512,7 @@ class AstBuilder(object):
         bases = []
         while True:
             token = self._get_next_token()
-            assert token.token_type == tokenize.NAME, token
+            assert_parse(token.token_type == tokenize.NAME, token)
             if token.name == 'virtual':
                 token = self._get_next_token()
             # TODO(nnorwitz): store kind of inheritance...maybe.
@@ -1525,21 +1531,22 @@ class AstBuilder(object):
                     pass
             base, next_token = self.get_name()
             bases_ast = self.converter.to_type(base)
-            assert len(bases_ast) == 1, bases_ast
+            assert_parse(len(bases_ast) == 1, bases_ast)
             bases.append(bases_ast[0])
-            assert next_token.token_type == tokenize.SYNTAX, next_token
+            assert_parse(next_token.token_type == tokenize.SYNTAX, next_token)
             if next_token.name == '{':
                 token = next_token
                 break
             # Support multiple inheritance.
-            assert next_token.name == ',', next_token
+            assert_parse(next_token.name == ',', next_token)
         return bases, token
 
     def _get_class(self, class_type, visibility, templated_types):
         class_name = None
         class_token = self._get_next_token()
         if class_token.token_type != tokenize.NAME:
-            assert class_token.token_type == tokenize.SYNTAX, class_token
+            assert_parse(class_token.token_type == tokenize.SYNTAX,
+                         class_token)
             token = class_token
         else:
             self._add_back_token(class_token)
@@ -1579,9 +1586,6 @@ class AstBuilder(object):
 
         body = None
         if token.token_type == tokenize.SYNTAX and token.name == '{':
-            assert token.token_type == tokenize.SYNTAX, token
-            assert token.name == '{', token
-
             ast = AstBuilder(self.get_scope(), self.filename, class_name,
                              visibility, self.namespace_stack)
             body = list(ast.generate())
@@ -1589,9 +1593,8 @@ class AstBuilder(object):
             if not self._handling_typedef:
                 token = self._get_next_token()
                 if token.token_type != tokenize.NAME:
-                    assert token.token_type == tokenize.SYNTAX, token
-                    if token.name != ';':
-                        raise ParseError(token)
+                    assert_parse(token.token_type == tokenize.SYNTAX, token)
+                    assert_parse(token.name == ';', token)
                 else:
                     new_class = class_type(class_token.start, class_token.end,
                                            class_name, bases, None,
@@ -1617,22 +1620,19 @@ class AstBuilder(object):
             name = token.name
             token = self._get_next_token()
         self.namespace_stack.append(name)
-        assert token.token_type == tokenize.SYNTAX, token
+        assert_parse(token.token_type == tokenize.SYNTAX, token)
         # Create an internal token that denotes when the namespace is complete.
         internal_token = tokenize.Token(_INTERNAL_TOKEN, _NAMESPACE_POP,
                                         None, None)
-        internal_token.whence = token.whence
         if token.name == '=':
             # TODO(nnorwitz): handle aliasing namespaces.
             name, next_token = self.get_name()
-            if next_token.name != ';':
-                raise ParseError(next_token)
+            assert_parse(next_token.name == ';', next_token)
             self._add_back_token(internal_token)
         else:
-            assert token.name == '{', token
+            assert_parse(token.name == '{', token)
             tokens = list(self.get_scope())
-            # Replace the trailing } with the internal namespace pop token.
-            tokens[-1] = internal_token
+            tokens.append(internal_token)
             # Handle namespace with nothing in it.
             self._add_back_tokens(tokens)
         return None
@@ -1686,7 +1686,7 @@ class AstBuilder(object):
 
     def handle_goto(self):
         tokens = self._get_tokens_up_to(tokenize.SYNTAX, ';')
-        assert len(tokens) == 1, unicode(tokens)
+        assert_parse(len(tokens) == 1, tokens)
         return Goto(tokens[0].start, tokens[0].end, tokens[0].name)
 
     def handle_try(self):
@@ -1726,3 +1726,9 @@ def builder_from_source(source, filename):
 
     """
     return AstBuilder(tokenize.get_tokens(source), filename)
+
+
+def assert_parse(value, message):
+    """Raise ParseError on token if value is False."""
+    if not value:
+        raise ParseError(message)
