@@ -60,6 +60,8 @@ UNUSED = 0
 USES_REFERENCE = 1
 USES_DECLARATION = 2
 
+DECLARATION_TYPES = (ast.Class, ast.Struct, ast.Enum, ast.Union)
+
 
 class Module(object):
 
@@ -89,9 +91,6 @@ def is_cpp_file(filename):
 
 class WarningHunter(object):
 
-    # Cache filename: ast_list
-    _module_cache = {}
-
     def __init__(self, filename, source, ast_list, include_paths):
         self.filename = filename
         self.normalized_filename = os.path.abspath(filename)
@@ -101,9 +100,7 @@ class WarningHunter(object):
         self.symbol_table = symbols.SymbolTable()
 
         self.metrics = metrics.Metrics(source)
-        self.warnings = []
-        if filename not in self._module_cache:
-            self._module_cache[filename] = Module(filename, ast_list)
+        self.warnings = set()
 
     def _add_warning(self, msg, node, filename=None):
         if filename is not None:
@@ -112,15 +109,14 @@ class WarningHunter(object):
             filename = self.filename
             src_metrics = self.metrics
         line_number = get_line_number(src_metrics, node)
-        self.warnings.append((filename, line_number, msg))
+        self.warnings.add((filename, line_number, msg))
 
     def show_warnings(self):
-        self.warnings.sort()
-        for filename, line_num, msg in self.warnings:
+        for filename, line_num, msg in sorted(self.warnings):
             if line_num == 0:
-                print('%s: %s' % (filename, msg))
+                print('{}: {}'.format(filename, msg))
             else:
-                print('%s:%d: %s' % (filename, line_num, msg))
+                print('{}:{}: {}'.format(filename, line_num, msg))
 
     def find_warnings(self):
         if is_header_file(self.filename):
@@ -134,41 +130,31 @@ class WarningHunter(object):
 
     def _get_module(self, node):
         filename = node.filename
-        if filename in self._module_cache:
-            # The cache survives across all instances, but the symbol table
-            # is per instance, so we need to make sure the symbol table
-            # is updated even if the module was in the cache.
-            module = self._module_cache[filename]
-            self._update_symbol_table(module)
-            return module
 
         (source, filename) = headers.read_source(
             filename,
-            include_paths=self.include_paths + [os.path.dirname(self.filename)]
+            include_paths=[os.path.dirname(self.filename)] + self.include_paths
         )
 
         if source is None:
             module = Module(filename, None)
-            msg = 'Unable to find %s' % filename
+            msg = "unable to find '{}'".format(filename)
             self._add_warning(msg, node)
         else:
             builder = ast.builder_from_source(source, filename)
             try:
                 module = Module(filename,
                                 [_f for _f in builder.generate() if _f])
-            except KeyboardInterrupt:
-                sys.exit(1)
-            except:
-                print('Exception while processing %s' % filename)
+            except ast.ParseError as error:
+                print("Exception while processing '{}': {}".format(filename,
+                                                                   error),
+                      file=sys.stderr)
                 module = Module(filename, None)
             else:
                 self._update_symbol_table(module)
-        self._module_cache[filename] = module
         return module
 
     def _read_and_parse_includes(self):
-        DECLARATION_TYPES = (ast.Class, ast.Struct, ast.Enum, ast.Union)
-
         # Map header-filename: (#include AST node, module).
         included_files = {}
         # Map declaration-name: AST node.
@@ -184,23 +170,39 @@ class WarningHunter(object):
 
         return included_files, forward_declarations
 
-    def _verify_includes(self, included_files):
+    def _verify_includes(self):
         """Read and parse all the #include'd files and warn about really stupid
         things that can be determined from the #include'd file name."""
         files_seen = {}
-        for filename, (node, module) in included_files.items():
-            normalized_filename = module.normalized_filename
-            if is_cpp_file(filename):
-                msg = 'should not #include C++ source file: %s' % filename
-                self._add_warning(msg, node)
-            if normalized_filename == self.normalized_filename:
-                self._add_warning('%s #includes itself' % filename, node)
-            if normalized_filename in files_seen:
-                include_node = files_seen[normalized_filename]
-                line_num = get_line_number(self.metrics, include_node)
-                msg = '%s already #included on line %d' % (filename, line_num)
-                self._add_warning(msg, node)
-            else:
+        for node in self.ast_list:
+            # Ignore #include <> files. Only handle #include "".
+            # Assume that <> are used for only basic C/C++ headers.
+            if isinstance(node, ast.Include) and not node.system:
+                module = self._get_module(node)
+                filename = module.normalized_filename
+
+                normalized_filename = module.normalized_filename
+
+                if is_cpp_file(filename):
+                    self._add_warning(
+                        "should not #include C++ source file '{}'".format(
+                            node.filename),
+                        node)
+
+                if normalized_filename == self.normalized_filename:
+                    self._add_warning(
+                        "'{}' #includes itself".format(node.filename),
+                        node)
+
+                if normalized_filename in files_seen:
+                    include_node = files_seen[normalized_filename]
+                    line_num = get_line_number(self.metrics, include_node)
+                    self._add_warning(
+                        "'{}' already #included on line {}".format(
+                            node.filename,
+                            line_num),
+                        node)
+
                 files_seen[normalized_filename] = node
 
     def _verify_include_files_used(self, file_uses, included_files):
@@ -209,24 +211,27 @@ class WarningHunter(object):
             if not use & USES_DECLARATION:
                 node, module = included_files[include_file]
                 if module.ast_list is not None:
-                    msg = module.filename + ' does not need to be #included'
+                    msg = "'{}' does not need to be #included".format(
+                        node.filename)
                     if use & USES_REFERENCE:
-                        msg += '. Use a forward declaration instead'
+                        msg += '; use a forward declaration instead'
                     self._add_warning(msg, node)
 
     def _verify_forward_declarations_used(self, forward_declarations,
                                           decl_uses, file_uses):
         """Find all the forward declarations that are not used."""
         for cls in forward_declarations:
-            if decl_uses[cls] == UNUSED:
-                node = forward_declarations[cls]
-                if cls in file_uses:
-                    msg = (
-                        "'{}' forward declared, "
-                        'but needs to be #included'.format(cls))
-                else:
+            if cls in file_uses:
+                if not decl_uses[cls] & USES_DECLARATION:
+                    node = forward_declarations[cls]
+                    msg = ("'{}' forward declared, "
+                           'but needs to be #included'.format(cls))
+                    self._add_warning(msg, node)
+            else:
+                if decl_uses[cls] == UNUSED:
+                    node = forward_declarations[cls]
                     msg = "'{}' not used".format(cls)
-                self._add_warning(msg, node)
+                    self._add_warning(msg, node)
 
     def _determine_uses(self, included_files, forward_declarations):
         """Set up the use type of each symbol."""
@@ -272,32 +277,26 @@ class WarningHunter(object):
                     name = '::'.join(namespace) + '::' + name
                 file_uses[name] = file_uses.get(name, 0) | USES_DECLARATION
                 return
-            if not file_use_node:
-                print(('Could not find #include file for %s in %s' %
-                      (name, namespace)))
-                return
+
             # TODO(nnorwitz): do proper check for ref/pointer/symbol.
             name = file_use_node[1].normalized_filename
             if name in file_uses:
                 file_uses[name] |= USES_DECLARATION
 
-        def _add_variable(node, name, namespace):
-            if not name:
-                # Assume that all the types without names are builtin.
-                return
+        def _add_variable(node, namespace):
             if node.reference or node.pointer:
-                _add_reference(name, namespace)
+                _add_reference(node.name, namespace)
             else:
-                _add_use(name, namespace)
+                _add_use(node.name, namespace)
             # This needs to recurse when the node is a templated type.
-            for n in node.templated_types or ():
-                _add_variable(n, n.name, namespace)
+            _add_template_use(node.name,
+                              node.templated_types,
+                              namespace)
 
         def _process_function(function):
             if function.return_type:
                 return_type = function.return_type
                 _add_variable(return_type,
-                              return_type.name,
                               function.namespace)
 
             templated_types = function.templated_types or ()
@@ -312,7 +311,7 @@ class WarningHunter(object):
                         # used.
                         _add_use(p.type.name, function.namespace)
                     else:
-                        _add_variable(p.type, p.type.name, function.namespace)
+                        _add_variable(p.type, function.namespace)
 
         def _process_function_body(function, namespace):
             iterator = iter(function.body)
@@ -346,10 +345,7 @@ class WarningHunter(object):
         while ast_seq:
             for node in ast_seq.pop():
                 if isinstance(node, ast.VariableDeclaration):
-                    _add_variable(node.type, node.type.name, node.namespace)
-                    _add_template_use(node.type.name,
-                                      node.type.templated_types,
-                                      node.namespace)
+                    _add_variable(node.type, node.namespace)
                 elif isinstance(node, ast.Function):
                     _process_function(node)
                     if node.body:
@@ -357,7 +353,10 @@ class WarningHunter(object):
                 elif isinstance(node, ast.Typedef):
                     alias = node.alias
                     if isinstance(alias, ast.Type):
-                        _add_use(alias.name, node.namespace)
+                        if alias.reference or alias.pointer:
+                            _add_reference(alias.name, node.namespace)
+                        else:
+                            _add_use(alias.name, node.namespace)
                         _add_template_use('<typedef>', alias.templated_types,
                                           node.namespace)
                 elif isinstance(node, ast.Friend):
@@ -377,7 +376,7 @@ class WarningHunter(object):
         included_files, forward_declarations = self._read_and_parse_includes()
         file_uses, decl_uses = self._determine_uses(included_files,
                                                     forward_declarations)
-        self._verify_includes(included_files)
+        self._verify_includes()
         self._verify_include_files_used(file_uses, included_files)
         self._verify_forward_declarations_used(forward_declarations, decl_uses,
                                                file_uses)
@@ -397,19 +396,24 @@ class WarningHunter(object):
                 # If the primary.filename == header.filename, it probably
                 # indicates an error elsewhere. It sucks to mask it,
                 # but false positives are worse.
-                if (primary_header and
-                        primary_header.filename != header.filename):
-                    msg = ('expected to find %s in %s, but found in %s' %
-                           (name, primary_header.filename, header.filename))
+                if primary_header:
+                    msg = ("expected to find '{}' in '{}', "
+                           "but found in '{}'".format(name,
+                                                      primary_header.filename,
+                                                      header.filename))
                     self._add_warning(msg, node)
                 break
         else:
             where = 'in any directly #included header'
             if primary_header:
-                where = ('in expected header ' + primary_header.filename +
-                         ' or any other directly #included header')
+                where = (
+                    "in expected header '{}'"
+                    ' or any other directly #included header'.format(
+                        primary_header.filename))
+
             if name != 'main':
-                self._add_warning('%s not found %s' % (name, where), node)
+                self._add_warning("'{}' not found {}".format(name, where),
+                                  node)
 
     def _check_public_functions(self, primary_header, all_headers):
         """Verify all the public functions are also declared in a header
@@ -456,7 +460,7 @@ class WarningHunter(object):
             if declared_only:
                 node = public_symbols[name]
                 if not node.templated_types:
-                    msg = '%s declared but not defined' % name
+                    msg = "'{}' declared but not defined".format(name)
                     self._add_warning(msg, node, primary_header.filename)
 
     def _get_primary_header(self, included_files):
@@ -473,13 +477,15 @@ class WarningHunter(object):
 
     def _find_source_warnings(self):
         included_files, forward_declarations = self._read_and_parse_includes()
-        if forward_declarations:
+        for node in forward_declarations.values():
             # TODO(nnorwitz): This really isn't a problem, but might
             # be something to warn against. I expect this will either
             # be configurable or removed in the future. But it's easy
             # to check for now.
-            msg = 'forward declarations not expected in source file'
-            self._add_warning(msg, list(forward_declarations.values())[0])
+            msg = (
+                "'{}' forward declaration not expected in source file".format(
+                    node.name))
+            self._add_warning(msg, node)
 
         # A primary header is optional. However, when looking up
         # defined methods in the source, always look in the
@@ -489,10 +495,21 @@ class WarningHunter(object):
         if not primary_header and not any(node for node in self.ast_list
                                           if isinstance(node, ast.Function) and
                                           node.name == 'main'):
-            msg = 'Unable to find header file with matching name'
-            self.warnings.append((self.filename, 0, msg))
+            msg = 'unable to find header file with matching name'
+            self.warnings.add((self.filename, 0, msg))
 
         self._check_public_functions(primary_header, included_files)
+        if primary_header:
+            includes = [
+                node.filename
+                for node in primary_header.ast_list
+                if (isinstance(node, ast.Include))
+            ]
+            for (node, _) in included_files.values():
+                if node.filename in includes:
+                    msg = "'{}' already #included in '{}'".format(
+                        node.filename, primary_header.filename)
+                    self._add_warning(msg, node)
 
         # TODO(nnorwitz): other warnings to add:
         #   * unused forward decls for variables (globals)/classes
